from __future__ import annotations
import os
import logging
import sqlite3
from datetime import datetime
from logging import Handler
from typing import TYPE_CHECKING
from geoprob_pipe.cmd_app.utils.misc import get_geoprob_pipe_version_number
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings
    from logging import LogRecord


class SQLiteHandler(Handler):

    def __init__(self, db_path: str):
        super().__init__()
        self.db_path = db_path
        self.app_version = get_geoprob_pipe_version_number()
        self.conn = sqlite3.connect(self.db_path)
        self.conn.execute("""
            CREATE TABLE IF NOT EXISTS logs (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                username TEXT,
                created TIMESTAMP,
                level TEXT,
                message TEXT,
                version TEXT,
                module TEXT,
                function TEXT,
                lineno INTEGER
            )
        """)
        self.conn.commit()

    def emit(self, record: LogRecord):
        created_dt = datetime.fromtimestamp(record.created).isoformat()
        username = os.getenv("USERNAME")
        self.conn.execute(
            "INSERT INTO logs (username, created, level, message, version, module, function, lineno) "
            "VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
            (
                username, created_dt, record.levelname, record.getMessage(),
                self.app_version, record.module, record.funcName, record.lineno
            ))
        self.conn.commit()


class ColorFormatter(logging.Formatter):

    COLORS = {
        logging.DEBUG: "\033[34m",     # blauw
        logging.INFO: "\033[32m",      # groen
        logging.WARNING: "\033[33m",   # geel
        logging.ERROR: "\033[31m",     # rood
        logging.CRITICAL: "\033[1;31m" # bold rood
    }

    RESET = "\033[0m"

    def format(self, record):
        color = self.COLORS.get(record.levelno, self.RESET)
        message = super().format(record)
        return f"{color}{message}{self.RESET}"



def setup_base_logging():
    """ Configure logging for the entire application at INFO-level. Call this once at application startup. Later on,
    when it is clear where the geopackage resides, a handler will be added that logs to the geopackage. """


    logger = logging.getLogger("geoprob-pipe")
    logger.setLevel(logging.DEBUG) # Capture DEBUG, but handlers filter output.

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)

    formatter = ColorFormatter("%(asctime)s - %(levelname)s - %(message)s")
    console_handler.setFormatter(formatter)

    if not logger.handlers:
        logger.addHandler(console_handler)

    logger.propagate = False


def enable_geopackage_logging(app_settings: ApplicationSettings):
    """ enable logging into the geopackage. Logging level is INFO, unless GeoProb-Pipe is run with the `debug`-flag,
    then level is DEBUG. """

    logger = logging.getLogger("geoprob-pipe")

    level = logging.INFO
    if app_settings.debug:
        level = logging.DEBUG

    sqlite_handler = SQLiteHandler(db_path=app_settings.geopackage_filepath)
    sqlite_handler.setLevel(level)

    logger.addHandler(sqlite_handler)

    logger.info("Logging into GeoProb-Pipe-file now active.")
    logger.debug("Test debug log message.")



from __future__ import annotations
import sqlite3
import os
from datetime import datetime
import pandas as pd
import geopandas as gpd
from plotly.graph_objects import Figure as PlotlyFigure
from geoprob_pipe.cmd_app.comparisons.beta_dumbbell import \
    dumbbell_beta, dumbbell_uplift, dumbbell_heave, dumbbell_piping
from geoprob_pipe.cmd_app.comparisons.beta_map import map_delta_beta_comparison, \
    map_ratio_beta_comparison


class ComparisonCollector:
    def __init__(self, geopackage_filepath_1: str, geopackage_filepath_2: str, export_dir: str):
        self.geopackage_filepath_1 = geopackage_filepath_1
        self.geopackage_filepath_2 = geopackage_filepath_2
        self.name_1 = os.path.basename(self.geopackage_filepath_1).replace(".geoprob_pipe.gpkg", "")
        self.name_2 = os.path.basename(self.geopackage_filepath_2).replace(".geoprob_pipe.gpkg", "")

        timestamp = datetime.now().strftime("%Y-%m-%d_%H%M")
        self.export_dir = os.path.join(
            export_dir, f"comparisons/{self.name_1}_{self.name_2}_{timestamp}"
            )
        os.makedirs(export_dir, exist_ok=True)

        # Placeholders
        self.df1_beta_scenarios: pd.DataFrame
        self.df1_beta_limit_states: pd.DataFrame
        self.df1_beta_uittredepunten: pd.DataFrame

        self.df2_beta_scenarios: pd.DataFrame
        self.df2_beta_limit_states: pd.DataFrame
        self.df2_beta_uittredepunten: pd.DataFrame

        self.gdf1_uittredepunten: gpd.GeoDataFrame
        self.gdf2_uittredepunten: gpd.GeoDataFrame

        # logic
        self._load_result_data_from_geopackage()
        self._load_uittredepunten_gdf()

    def _load_result_data_from_geopackage(self):

        conn_1 = sqlite3.connect(self.geopackage_filepath_1)
        conn_2 = sqlite3.connect(self.geopackage_filepath_2)

        self.df1_beta_limit_states = pd.read_sql(
            "SELECT * FROM beta_limit_states;", conn_1
            )
        self.df2_beta_limit_states = pd.read_sql(
            "SELECT * FROM beta_limit_states;", conn_2
            )
        if len(self.df1_beta_limit_states) != len(self.df2_beta_limit_states):
            raise ValueError("De beta_limit_states tables hebben niet hetzelfde formaat")

        self.df1_beta_scenarios = pd.read_sql(
            "SELECT * FROM beta_scenarios;", conn_1
            )
        self.df2_beta_scenarios = pd.read_sql(
            "SELECT * FROM beta_scenarios;", conn_2
            )
        if len(self.df1_beta_scenarios) != len(self.df2_beta_scenarios):
            raise ValueError("De beta_scenario tables hebben niet hetzelfde formaat")

        self.df1_beta_uittredepunten = pd.read_sql(
            "SELECT * FROM beta_uittredepunten;", conn_1
            )
        self.df2_beta_uittredepunten = pd.read_sql(
            "SELECT * FROM beta_uittredepunten;", conn_2
            )
        if len(self.df1_beta_uittredepunten) != len(self.df2_beta_uittredepunten):
            raise ValueError("De beta_uittredepunten tables hebben niet hetzelfde formaat")

        conn_1.close()
        conn_2.close()

    def _load_uittredepunten_gdf(self):
        self.gdf1_uittredepunten = gpd.read_file(
            self.geopackage_filepath_1,
            layer="beta_uittredepunten"
            )
        self.gdf2_uittredepunten = gpd.read_file(
            self.geopackage_filepath_2,
            layer="beta_uittredepunten"
            )
        if len(self.gdf1_uittredepunten) != len(self.gdf2_uittredepunten):
            raise ValueError("De beta_uittredepunten tables hebben niet hetzelfde formaat")
        if set(self.gdf1_uittredepunten.geometry) != set(self.gdf2_uittredepunten.geometry):
            raise ValueError("De twee sets uittredepunten hebben afwijkende geometry")

    def dumbbell_beta(self, export: bool = False) -> PlotlyFigure:
        return dumbbell_beta(self, export)

    def dumbbell_uplift(self, export: bool = False) -> list[PlotlyFigure]:
        return dumbbell_uplift(self, export)

    def dumbbell_heave(self, export: bool = False) -> list[PlotlyFigure]:
        return dumbbell_heave(self, export)

    def dumbbell_piping(self, export: bool = False) -> list[PlotlyFigure]:
        return dumbbell_piping(self, export)

    def map_delta_beta_comparison(self, export: bool = False) -> PlotlyFigure:
        return map_delta_beta_comparison(self, export)

    def map_ratio_beta_comparison(self, export: bool = False) -> PlotlyFigure:
        return map_ratio_beta_comparison(self, export)

    def create_and_export_figures(self):
        dumbbell_beta(self, export=True)
        dumbbell_uplift(self, export=True)
        dumbbell_heave(self, export=True)
        dumbbell_piping(self, export=True)
        map_delta_beta_comparison(self, export=True)
        map_ratio_beta_comparison(self, export=True)



from __future__ import annotations
from typing import TYPE_CHECKING
from geoprob_pipe.visualizations.graphs import Graphs
from geoprob_pipe.visualizations.maps import Maps
import os
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


class Visualizations:

    def __init__(self, app_obj: GeoProbPipe):
        self.geoprob_pipe = app_obj
        self.graphs = Graphs(app_obj)
        self.maps = Maps(app_obj)

    @property
    def export_dir(self) -> str:
        path = os.path.join(
            self.geoprob_pipe.input_data.app_settings.workspace_dir,
            "exports",
            self.geoprob_pipe.input_data.app_settings.datetime_stamp,
            "visualizations"
        )
        os.makedirs(path, exist_ok=True)
        return str(path)

    def export_visualizations(self):
        self.graphs.export_graphs()
        self.maps.export_maps()



from __future__ import annotations
from geopandas import GeoDataFrame, read_file
import scipy.stats as sct
from pandas import read_sql, DataFrame
from probabilistic_library import FragilityValue
import pydra_core as pydra
from shapely import Point
from typing import Optional, TYPE_CHECKING, List
import sqlite3
from geoprob_pipe.input_data.traject_normering import TrajectNormering
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


class Vak:

    def __init__(self, app_settings: ApplicationSettings, vak_id: int):
        self.app_settings: ApplicationSettings = app_settings
        self.vak_id = vak_id


class Uittredepunten:

    def __init__(self, app_settings: ApplicationSettings):
        self.app_settings: ApplicationSettings = app_settings
        self.gdf: GeoDataFrame = read_file(app_settings.geopackage_filepath,
                                           layer="uittredepunten")

    def uittredepunt(self, uittredepunt_id: int):
        row = self.gdf.loc[
            self.gdf['uittredepunt_id'] == uittredepunt_id
            ].iloc[0]
        kwargs = row.to_dict()
        return Uittredepunt(self.app_settings, **kwargs)


class Uittredepunt:

    def __init__(
            self, app_settings: ApplicationSettings, geometry: Point,
            uittredepunt_id: int, vak_id: int, **_):
        self.app_settings: ApplicationSettings = app_settings
        self.geometry: Point = geometry
        self.uittredepunt_id: int = uittredepunt_id
        self.vak_id: int = vak_id

    @property
    def vak(self) -> Vak:
        return Vak(app_settings=self.app_settings, vak_id=self.vak_id)


class Scenarios:

    def __init__(self, app_settings: ApplicationSettings):
        self.app_settings: ApplicationSettings = app_settings
        self.df = self._query_scenarios()

    def _query_scenarios(self) -> DataFrame:
        conn = sqlite3.connect(self.app_settings.geopackage_filepath)
        df_scenario_invoer = read_sql("SELECT * FROM scenario_invoer;", conn)
        df_scenario_invoer = df_scenario_invoer[["vak_id", "naam", "kans"]]
        conn.close()
        return df_scenario_invoer

    def scenario_kans(self, vak_id: int, scenario_naam: str) -> float:
        kans: float = self.df[
            (self.df['vak_id'] == vak_id) &
            (self.df['naam'] == scenario_naam)
        ]['kans'].iloc[0]
        return kans


class HydraNLData:

    def __init__(self, app_settings: ApplicationSettings):
        self.app_settings: ApplicationSettings = app_settings
        self.gdf_locations: GeoDataFrame = read_file(
            app_settings.geopackage_filepath, layer="hrd_locaties")

    def hrd_fragility_values(self, ref: str) -> List[FragilityValue]:

        # Read from geopackage
        conn = sqlite3.connect(self.app_settings.geopackage_filepath)
        df_frag_line = read_sql(
            sql=f"SELECT * FROM fragility_values_invoer_hrd WHERE fragility_values_ref = '{ref}' AND kans < 1.0;",
            con=conn)
        conn.close()

        # Filter beta > 8 (probabilistic library cannot work with that in 26.1.1)
        df_frag_line['beta'] = -sct.norm.ppf(df_frag_line['kans'])
        df_frag_line = df_frag_line[df_frag_line['beta'] < 8.0].copy(deep=True)
        df_frag_line = df_frag_line.drop(columns=["beta"])

        # Construct Fragility Values
        df_frag_line = df_frag_line.sort_values(by=["waarde"])
        frag_points = []
        for _, row in df_frag_line.iterrows():
            fc = FragilityValue()
            fc.x = row["waarde"]
            fc.probability_of_failure = row["kans"]
            frag_points.append(fc)

        return frag_points

    # noinspection PyUnresolvedReferences
    def hrd_frequency_line(
            self, ref: str
            ) -> pydra.core.datamodels.frequency_line.FrequencyLine:
        hrd_fragility_values = self.hrd_fragility_values(ref=ref)
        level = [item.x for item in hrd_fragility_values]
        exceedance_frequency = [item.probability_of_failure
                                for item in hrd_fragility_values]
        # noinspection PyUnresolvedReferences
        return pydra.core.datamodels.frequency_line.FrequencyLine(
            level=level, exceedance_frequency=exceedance_frequency)


class Vakken:

    def __init__(self, app_settings: ApplicationSettings):
        self.app_settings: ApplicationSettings = app_settings
        self.gdf: GeoDataFrame = read_file(app_settings.geopackage_filepath,
                                           layer="vakindeling")


class InputData:
    """ Subclass to group input data of vakken, uittredepunten and
    ondergrondscenarios. Data is retrieved from the input Excel-file.
    """

    def __init__(
            self,
            app_settings: ApplicationSettings
            # workspace: Workspace,
    ):

        self.app_settings: ApplicationSettings = app_settings

        # Traject-data
        self._traject_normering: Optional[TrajectNormering] = None

        self.uittredepunten = Uittredepunten(self.app_settings)
        self.scenarios = Scenarios(self.app_settings)
        self.vakken = Vakken(self.app_settings)
        self.hydra_nl_data = HydraNLData(self.app_settings)

    @property
    def geohydrologisch_model(self) -> str:
        conn = sqlite3.connect(self.app_settings.geopackage_filepath)
        cursor = conn.cursor()
        cursor.execute("""
            SELECT geoprob_pipe_metadata."values"
            FROM geoprob_pipe_metadata
            WHERE metadata_type='geohydrologisch_model';
        """)
        result = cursor.fetchone()
        if not result:
            raise ValueError
        model_string = result[0]
        conn.close()
        return model_string

    @property
    def traject_normering(self):
        if self._traject_normering is None:
            self._traject_normering = TrajectNormering(
                hrd_path=self.app_settings.hrd_file_path)
        return self._traject_normering



from __future__ import annotations
from InquirerPy import inquirer
from typing import TYPE_CHECKING
from rich.console import Console
from rich.panel import Panel
from geoprob_pipe.utils import clear_terminal
import os
from pandas import DataFrame
from datetime import datetime
from pathlib import Path
import sys
from importlib.metadata import distributions
from geopandas import GeoDataFrame
from geoprob_pipe.cmd_app.utils.misc import get_geoprob_pipe_version_number
from geoprob_pipe.cmd_app.comparisons.start_comparison import start_comparison
from geoprob_pipe.calculations.systems.single_calc import (
    EXAMPLE_SCRIPT_REPRODUCING_SINGLE_CALCULATION, EXPLANATION_REPRODUCING_SINGLE_CALCULATION)
import logging
from packaging.version import Version
from geoprob_pipe.utils.validation_messages import BColors
import sqlite3
from typing import Optional
import ast
from geoprob_pipe.utils.loggers import enable_geopackage_logging
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings



logger = logging.getLogger("geoprob-pipe")


def created_project(app_settings: ApplicationSettings) -> bool:

    choices_list = [
        "Bestaand project openen",
        "Nieuw project starten",
        "Twee projectbestanden vergelijken",
        "Inspecteer een enkele berekening",
        "Applicatie afsluiten"]
    choice = inquirer.select(
        message="Wil je verder gaan met een bestaand project, "
                "een nieuw project starten, "
                "of twee project bestanden vergelijken?",
        choices=choices_list,
        default=choices_list[0],
    ).execute()

    if choice == choices_list[0]:
        specify_path_to_existing_project(app_settings)
        enable_geopackage_logging(app_settings=app_settings)
        return True

    elif choice == choices_list[1]:
        specify_dir_for_new_project(app_settings)
        enable_geopackage_logging(app_settings=app_settings)
        return True

    elif choice == choices_list[2]:
        start_comparison()
        sys.exit("Applicatie afgesloten")

    elif choice == choices_list[3]:
        clear_terminal()
        console = Console()
        console.print(Panel(
            EXPLANATION_REPRODUCING_SINGLE_CALCULATION,
            title=f"Inspecteer een enkele berekening".upper(),
            title_align="left",
            border_style="bright_blue",
            padding=(0, 2)))
        print(EXAMPLE_SCRIPT_REPRODUCING_SINGLE_CALCULATION)
        sys.exit("Applicatie afgesloten")

    return False


def _get_file_geoprob_pipe_version(file_path: str) -> Optional[str]:
    conn = sqlite3.connect(file_path)
    cursor = conn.cursor()
    cursor.execute(f"SELECT * FROM geoprob_pipe_metadata WHERE metadata_type = 'pip_freeze' LIMIT 1;")
    row = cursor.fetchone()
    dict_item = ast.literal_eval(row[2])

    # Check if created with development environment
    if "geoprob_pipe" not in dict_item:
        return None

    return dict_item["geoprob_pipe"]


def _compare_versions_and_possibly_warn(installed_version: str, file_version: str):

    v_installed = Version(installed_version)
    v_file = Version(file_version)

    version_diff_msg = (
        f"Er is een GeoProb-Pipe versie verschil tussen de installatie ({v_installed}) en het bestand ({v_file}).")
    reinstall_msg = (
        f"\n"
        f"Wil je de versies liever gelijk trekken? Dit doe je door eerst de applicatie af te sluiten en vervolgens het \n"
        f"commando `pip install geoprob-pipe==versie_nummer` uit te voeren. Waarbij je `versie_nummer` vervangt met het \n"
        f"nummer van de gewenste versie")

    if v_installed.major != v_file.major:
        print(f"\n{BColors.FAIL}SEVERE WARNING:\n"
              f"{version_diff_msg}\n"
              f"Dit is een versie verschil met major changes. Dit is een risico. Het kan goed zijn dat de applicatie niet "
              f"compatibel is met je bestand. \n"
              f"{reinstall_msg}"
              f"{BColors.ENDC}\n")
    elif v_installed.minor != v_file.minor:
        print(f"\n{BColors.WARNING}WARNING:\n"
              f"{version_diff_msg}\n"
              f"Dit is een versie verschil met minor changes. Dit kan mogelijk een probleem zijn tijdens het gebruik van "
              f"de applicatie. \n"
              f"{reinstall_msg}"
              f"{BColors.ENDC}\n")
    elif v_installed.micro != v_file.micro:
        print(f"\n{BColors.OKBLUE}"
              f"{version_diff_msg}\n"
              f"Dit verschil is slechts op patch-niveau. Dit is waarschijnlijk geen probleem tijdens het gebruik van "
              f"de applicatie. \n"
              f"{reinstall_msg}"
              f"{BColors.ENDC}\n")


def _possibly_warn_if_version_difference(file_path: str):

    installed_app_version = get_geoprob_pipe_version_number()
    if installed_app_version == "DEV":
        return  # We do not warn if application is run in development environment.

    file_app_version = _get_file_geoprob_pipe_version(file_path=file_path)
    if file_app_version is None:
        return  # Version was not found what probably means it was made in a development environment.

    _compare_versions_and_possibly_warn(installed_version=installed_app_version, file_version=file_app_version)


def specify_path_to_existing_project(app_settings: ApplicationSettings):
    filepath: Optional[str] = None
    filepath_is_valid = False
    while filepath_is_valid is False:
        filepath: str = inquirer.text(
            message="Specificeer het volledige bestandspad naar het .geoprob_pipe.gpkg-bestand.",
        ).execute()

        filepath = filepath.replace('"', '')

        if not filepath.endswith(".geoprob_pipe.gpkg"):
            print(BColors.WARNING, f"Het bestand moet een .geoprob_pipe.gpkg-bestand zijn. Jouw invoer "
                                   f"{os.path.basename(filepath)} eindigt niet op deze extensie.", BColors.ENDC)
            continue
        if not os.path.exists(filepath):
            print(BColors.WARNING, f"Het opgegeven bestandspad bestaat niet.", BColors.ENDC)
            continue

        filepath_is_valid = True

    app_settings.workspace_dir = os.path.dirname(filepath)
    app_settings.geopackage_filename = os.path.basename(filepath)

    _possibly_warn_if_version_difference(file_path=app_settings.geopackage_filepath)


def specify_dir_for_new_project(app_settings: ApplicationSettings):
    workspace_dir: Optional[str] = None
    workspace_dir_is_valid = False
    while workspace_dir_is_valid is False:
        workspace_dir: str = inquirer.text(
            message="Specificeer het volledige pad naar de map waar je het GeoProb-Pipe-bestand wilt opslaan.",
        ).execute()
        workspace_dir = workspace_dir.replace('"', '')

        if not os.path.exists(workspace_dir):
            print(BColors.WARNING, f"De opgegeven map bestaat niet.", BColors.ENDC)
            continue
        if not os.path.isdir(workspace_dir):
            print(BColors.WARNING, f"De opgegeven locatie is geen map.", BColors.ENDC)
            continue

        workspace_dir_is_valid = True

    app_settings.workspace_dir = workspace_dir

    # Continue questionnaire
    specify_project_filename(app_settings)


def specify_project_filename(app_settings: ApplicationSettings):
    filename: Optional[str] = None
    filename_is_valid = False
    while filename_is_valid is False:
        project_name: str = inquirer.text(
            message="Specificeer een bestandsnaam voor het project. "
                    "Het bestand wordt opgeslagen met een .geoprob_pipe.gpkg-extensie.",
        ).execute()

        filename = f"{project_name}.geoprob_pipe.gpkg"
        filepath = os.path.join(app_settings.workspace_dir, filename)

        if os.path.exists(filepath):
            print(BColors.WARNING, f"De opgegeven bestandsnaam bestaat al. "
                                   f"Kies een andere naam. "
                                   f"Je gaf het volgden op:\n"
                                   f" {filepath}", BColors.ENDC)
            continue

        filename_is_valid = True

    app_settings.geopackage_filename = filename

    create_geopackage_file(app_settings)


def create_geopackage_file(app_settings: ApplicationSettings):

    df = DataFrame({
        "metadata_type": ["created_by", "created_datetime", "application_version", "python_version", "pip_freeze"],
        "values": [
            os.getenv("USERNAME"),
            datetime.now(),
            get_geoprob_pipe_version_number(),
            f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            {dist.metadata["Name"]: dist.version for dist in distributions()},
        ]
    })

    gdf = GeoDataFrame(df, geometry=None)

    gdf.to_file(Path(app_settings.geopackage_filepath), layer="geoprob_pipe_metadata", driver="GPKG")
    print(BColors.OKBLUE, f"✅  Project-bestand aangemaakt op locatie:\n {app_settings.geopackage_filepath}", BColors.ENDC)
